import dataclasses
from typing import Any, List, Optional, Tuple

from jax import numpy as jnp
from typing_extensions import TypeGuard, get_type_hints

ExpectedShape = Tuple[Any, ...]


def _is_expected_shape(shape: Any) -> TypeGuard[ExpectedShape]:
    """Returns `True` is a tuple of integers + potentially an Ellipsis."""
    return isinstance(shape, tuple) and all(
        map(lambda x: isinstance(x, int) or x is ..., shape)
    )


# Some dtype superclasses that result in a warning when we attempt a jnp.dtype() on them.
_dtype_set = {jnp.integer, jnp.signedinteger, jnp.floating, jnp.inexact}


def _is_dtype(dtype: Any) -> bool:
    """Returns `True` is `dtype` is a valid datatype."""

    if dtype in _dtype_set:
        return True
    try:
        jnp.dtype(dtype)
        return True
    except TypeError:
        return False


class EnforcedAnnotationsMixin:
    """Mixin for dataclasses containing arrays that are annotated with expected shapes
    or types that can be checked at runtime.

    Runs input validation on instantiation and provides a single helper,
    `get_batch_axes()`, that returns common batch axes.

    First, we import `Annotated`:

        # Python <=3.8
        from typing_extensions import Annotated

        # Python 3.9
        from typing import Annotated

    Example of an annotated fields that must have shape (*, 50, 150, 3) and (*, 10),
    where the batch axes must be shared:

        image: Annotated[jnp.ndarray, (..., 50, 150, 3)]
        label: Annotated[jnp.ndarray, (..., 10,)]

    Fields that must be floats and integers respectively:

        image: Annotated[jnp.ndarray, jnp.floating] # or jnp.float32, jnp.float64, etc
        label: Annotated[jnp.ndarray, jnp.integer] # or jnp.uint8, jnp.uint32, etc

    Fields with both shape and type constraints:

        image: Annotated[jnp.ndarray, (..., 50, 150, 3), jnp.floating]
        label: Annotated[jnp.ndarray, (..., 10,), jnp.integer]

    Where the annotations are order-invariant and both optional.
    """

    def __post_init__(self):
        """Validate after construction.

        We raise assertion errors in only two scenarios:
        - A field has a dtype that's not a subtype of the annotated dtype.
        - A field has a shape that's not consistent with the annotated shape."""

        assert dataclasses.is_dataclass(self)

        hint_from_name = get_type_hints(type(self), include_extras=True)
        batch_axes: Optional[Tuple[int, ...]] = None

        # Batch axes for child/nested elements.
        child_batch_axes_list: List[Tuple[int, ...]] = []

        # For each field...
        for field in dataclasses.fields(self):
            type_hint = hint_from_name[field.name]
            value = self.__getattribute__(field.name)

            if isinstance(value, EnforcedAnnotationsMixin):
                child_batch_axes = getattr(value, "__batch_axes__")
                if child_batch_axes is not None:
                    child_batch_axes_list.append(child_batch_axes)
                continue

            # Check for metadata from `typing.Annotated` value! Skip if no annotation.
            if not hasattr(type_hint, "__metadata__"):
                continue
            metadata: Tuple[Any, ...] = type_hint.__metadata__
            assert (
                len(metadata) <= 2
            ), "We expect <= 2 metadata items; only shape and dtype are expected."

            # Check data type.
            metadata_dtype = tuple(filter(_is_dtype, metadata))
            if len(metadata_dtype) > 0 and hasattr(value, "dtype"):
                (dtype,) = metadata_dtype
                assert jnp.issubdtype(
                    value.dtype, dtype
                ), f"Mismatched dtype, expected {dtype} but got {value.dtype}."

            # Shape checks.
            metadata_shape = tuple(filter(_is_expected_shape, metadata))
            shape: Optional[Tuple[int, ...]] = None
            if isinstance(value, (int, float)):
                shape = ()
            elif hasattr(value, "shape"):
                shape = value.shape
            if len(metadata_shape) > 0 and shape is not None:
                # Get expected shape, sans batch axes.
                (expected_shape,) = metadata_shape
                field_batch_axes = _check_batch_axes(shape, expected_shape)
                if batch_axes is None:
                    batch_axes = field_batch_axes
                else:
                    assert (
                        batch_axes == field_batch_axes
                    ), f"Batch axis mismatch: {batch_axes} and {field_batch_axes}."

        # Check child batch axes: any batch axes present in the parent should be present
        # in the children as well.
        if batch_axes is not None:
            for child_batch_axes in child_batch_axes_list:
                assert (
                    len(child_batch_axes) >= len(batch_axes)
                    and child_batch_axes[: len(batch_axes)] == batch_axes
                ), (
                    f"Child batch axes {child_batch_axes} don't match parent axes"
                    f" {batch_axes}."
                )

        object.__setattr__(self, "__batch_axes__", batch_axes)

    def get_batch_axes(self) -> Tuple[int, ...]:
        """Return any leading batch axes (which should be shared across all contained
        arrays)."""

        batch_axes = getattr(self, "__batch_axes__")
        assert batch_axes is not None
        return batch_axes


def _check_batch_axes(
    shape: Tuple[int, ...],
    expected_shape: ExpectedShape,
) -> Tuple[int, ...]:
    # By default, assume batch axes are at start of
    if ... not in expected_shape:
        expected_shape = (...,) + expected_shape

    assert expected_shape.count(...) == 1
    batch_index = expected_shape.index(...)

    # Actual shape should be expected shape prefixed by some batch axes.
    if len(expected_shape) > 1:
        expected_prefix = expected_shape[:batch_index]
        expected_suffix = expected_shape[batch_index + 1 :]

        prefix = shape[: len(expected_prefix)]
        suffix = shape[-len(expected_suffix) :]

        shape_error = (
            "Shape did not match annotation: expected"
            f" ({','.join(map(str,expected_prefix + ('*',) + expected_suffix))}) but"
            f" got {shape}."
        )
        assert suffix == expected_suffix, shape_error
        assert prefix == expected_prefix, shape_error
        batch_axes = shape[len(expected_prefix) : -len(expected_suffix)]
    else:
        batch_axes = shape

    return batch_axes



import dataclasses
from typing import Dict, List, Optional, Type, TypeVar

import jax

try:
    # Attempt to import flax for serialization. The exception handling lets us drop
    # flax from our dependencies.
    from flax import serialization
except ImportError:
    serialization = None  # type: ignore

from . import _copy_and_mutate

T = TypeVar("T")


FIELD_METADATA_STATIC_MARKER = "__jax_dataclasses_static_field__"


def pytree_dataclass(cls: Optional[Type] = None, **kwargs):
    """Substitute for dataclasses.dataclass, which also registers dataclasses as
    PyTrees."""

    def wrap(cls):
        return _register_pytree_dataclass(dataclasses.dataclass(cls, **kwargs))

    if "frozen" in kwargs:
        assert kwargs["frozen"] is True, "Pytree dataclasses can only be frozen!"
    kwargs["frozen"] = True

    if cls is None:
        return wrap
    else:
        return wrap(cls)


def static_field(*args, **kwargs):
    """Substitute for dataclasses.field, which also marks a field as static."""

    kwargs["metadata"] = kwargs.get("metadata", {})
    kwargs["metadata"][FIELD_METADATA_STATIC_MARKER] = True

    return dataclasses.field(*args, **kwargs)


def _register_pytree_dataclass(cls: Type[T]) -> Type[T]:
    """Register a dataclass as a flax-serializable pytree container.

    Args:
        cls (Type[T]): Dataclass to wrap.
    """

    assert dataclasses.is_dataclass(cls)

    # Determine which fields are static and part of the treedef, and which should be
    # registered as child nodes.
    child_node_field_names: List[str] = []
    static_field_names: List[str] = []

    for field in dataclasses.fields(cls):
        if not field.init:
            continue
        if field.metadata.get(FIELD_METADATA_STATIC_MARKER, False):
            static_field_names.append(field.name)
        else:
            child_node_field_names.append(field.name)

    # Define flatten, unflatten operations: this simple converts our dataclass to a list
    # of fields.
    def _flatten(obj):
        children = tuple(getattr(obj, key) for key in child_node_field_names)
        treedef = tuple(getattr(obj, key) for key in static_field_names)
        return children, treedef

    def _unflatten(treedef, children):
        return cls(
            **dict(zip(child_node_field_names, children)),
            **{key: tdef for key, tdef in zip(static_field_names, treedef)},
        )

    jax.tree_util.register_pytree_node(cls, _flatten, _unflatten)

    # Serialization: this is mostly copied from `flax.struct.dataclass`.
    if serialization is not None:

        def _to_state_dict(x: T):
            state_dict = {
                name: serialization.to_state_dict(getattr(x, name))
                for name in child_node_field_names
            }
            return state_dict

        def _from_state_dict(x: T, state: Dict):
            # Copy the state so we can pop the restored fields.
            state = state.copy()
            updates = {}
            for name in child_node_field_names:
                if name not in state:
                    raise ValueError(
                        f"Missing field {name} in state dict while restoring"
                        f" an instance of {cls.__name__}"
                    )
                value = getattr(x, name)
                value_state = state.pop(name)
                updates[name] = serialization.from_state_dict(value, value_state)
            if state:
                names = ",".join(state.keys())
                raise ValueError(
                    f'Unknown field(s) "{names}" in state dict while'
                    f" restoring an instance of {cls.__name__}"
                )
            return dataclasses.replace(x, **updates)

        serialization.register_serialization_state(
            cls, _to_state_dict, _from_state_dict
        )

    # Custom frozen dataclass implementation
    cls.__mutability__ = _copy_and_mutate._Mutability.FROZEN  # type: ignore
    cls.__setattr__ = _copy_and_mutate._new_setattr  # type: ignore

    return cls



import contextlib
import dataclasses
import enum
from typing import Any, ContextManager, Sequence, Set, TypeVar

import jax
from jax import numpy as jnp
from jax._src.tree_util import _registry  # Dangerous!

T = TypeVar("T")


class _Mutability(enum.Enum):
    FROZEN = enum.auto()
    MUTABLE = enum.auto()
    MUTABLE_NO_VALIDATION = enum.auto()


def _tree_children(container: Any) -> Sequence[Any]:
    """Grab child nodes of a pytree. This would ideally be implemented using the pytree
    registry."""

    registry_entry = _registry.get(type(container))
    if registry_entry is not None:
        children, _metadata = registry_entry.to_iter(container)
        return list(children)
    return []


def _mark_mutable(obj: Any, mutable: _Mutability, visited: Set[Any]) -> None:
    """Recursively freeze or unfreeze dataclasses in a structure.
    Currently only supports tuples, lists, dictionaries, dataclasses."""

    # Skip objects we've already visited. This avoids redundancies when there are
    # identical branches in our pytree, but will also help prevent infinite looping from
    # cycles.
    if id(obj) in visited:
        return
    visited.add(id(obj))

    if dataclasses.is_dataclass(obj):
        object.__setattr__(obj, "__mutability__", mutable)

    for child in _tree_children(obj):
        _mark_mutable(child, mutable, visited)


def copy_and_mutate(obj: T, validate: bool = True) -> ContextManager[T]:
    """Context manager that copies a pytree and allows for temporary mutations to
    contained dataclasses. Optionally validates that treedefs, array shapes, and dtypes
    are not changed."""

    # Inner function helps with static typing!
    def _replace_context(obj: T):
        # Make a copy of the input object.
        obj_copy = jax.tree_map(lambda leaf: leaf, obj)

        # Mark it as mutable.
        _mark_mutable(
            obj_copy,
            mutable=_Mutability.MUTABLE
            if validate
            else _Mutability.MUTABLE_NO_VALIDATION,
            visited=set(),
        )

        # Yield.
        yield obj_copy

        # When done, mark as immutable again.
        _mark_mutable(
            obj_copy,
            mutable=_Mutability.FROZEN,
            visited=set(),
        )

    return contextlib.contextmanager(_replace_context)(obj)


def _unify_floats(dtype):
    if dtype == jnp.float64:
        return jnp.float32
    else:
        return dtype


def _new_setattr(self, name: str, value: Any):
    if self.__mutability__ == _Mutability.MUTABLE:
        # Validate changes.
        current_value = getattr(self, name)

        # Make sure tree structure is unchanged.
        assert jax.tree_structure(value) == jax.tree_structure(
            current_value
        ), "Mismatched tree structure!"

        # Check leaf shapes.
        new_shapes = tuple(
            leaf.shape for leaf in jax.tree_leaves(value) if hasattr(leaf, "shape")
        )
        cur_shapes = tuple(
            leaf.shape
            for leaf in jax.tree_leaves(current_value)
            if hasattr(leaf, "shape")
        )
        assert (
            new_shapes == cur_shapes
        ), f"Shape error: {new_shapes} does not match {cur_shapes}!"

        # Check leaf dtypes.
        new_dtypes = tuple(
            _unify_floats(leaf.dtype)
            for leaf in jax.tree_leaves(value)
            if hasattr(leaf, "dtype")
        )
        cur_dtypes = tuple(
            _unify_floats(leaf.dtype)
            for leaf in jax.tree_leaves(current_value)
            if hasattr(leaf, "dtype")
        )
        assert (
            new_dtypes == cur_dtypes
        ), f"Type error: {new_dtypes} does not match {cur_dtypes}!"

        object.__setattr__(self, name, value)

    elif self.__mutability__ == _Mutability.MUTABLE_NO_VALIDATION:
        # Make changes without validation.
        object.__setattr__(self, name, value)

    elif self.__mutability__ == _Mutability.FROZEN:
        raise dataclasses.FrozenInstanceError(
            "Dataclass registered as pytree is immutable!"
        )

    else:
        assert False
